# Print out, for each class of sextic twist parameter, the
# conjectural rank parity of a curve in that class which satisfies
# legendre(-3, T(d)) == 1.  Curves in the class which have
# legendre(-3, T(d)) == -1 have the opposite parity.
# This function puts the results in a big table, with units varying
# across rows and powers of 2,3 varying across columns.
#
# Table in the paper comes looking at the entries where at most one
# of {2, 3} have nontrivial residue.
def classifySexticTwistsOverQTable(sign = 1):
    K.<y> = NumberField(x-1)
    R = K.maximal_order()
    
    # for K = QQ, m_2(6) = m_2(2) = 3, m_3(6) = m_3(3) = 2.5 -> 3
    #units2ToM2Mod6 = getResiduesModPower(K.ideal(2)^3, 6)
    units2ToM2Mod6 = [1, 3, 5, 7]
    #units3ToM3Mod6 = getResiduesModPower(K.ideal(3)^3, 6)
    # Above returns [1, 23, 16, 17, 13, 2], but this is equivalent mod
    # 6th powers to: 1, 2, 4=13, 5=23, 8=17, 16
    units3ToM3Mod6 = [1, 2, 4, 5, 8, 16]
    print("2 Residues: " + str(units2ToM2Mod6))
    print("3 Residues: " + str(units3ToM3Mod6))
    
    for ord2 in range(0, 6):
        for ord3 in range(0, 6):
            sys.stdout.write(str(ord2) + "&" + str(ord3) + "&& ")
            sys.stdout.flush()
            for unit2Residue in units2ToM2Mod6:
                for unit3Residue in units3ToM3Mod6:
                    twistParam = findRepresentative(R, ord2, ord3, \
                            unit2Residue, unit3Residue, sign)
                    if (twistParam == 0): continue
                    legendreCond = legendreCondition(K, twistParam)
                    ETwist = EllipticCurve(QQ, [0, twistParam])
                    #try:
                    #    selmerRankParity = ETwist.simon_two_descent()[1] % 2
                    #    if legendreCond == 1:
                    #        print selmerRankParity
                    #    else:
                    #        print (1 - selmerRankParity)
                    #    #print selmerRankParity
                    #except:
                    #    print "Error:", sys.exc_info()[1]
                    rankParity = getConjecturalRankParity(ETwist)
                    #rankParity = ETwist.analytic_rank(algorithm='rubinstein') % 2
                    if legendreCond == 1:
                        sys.stdout.write(str(rankParity))
                    else:
                        sys.stdout.write(str(1 - rankParity))
                    if not (unit2Residue == units2ToM2Mod6[-1] and \
                            unit3Residue == units3ToM3Mod6[-1]):
                        sys.stdout.write("&")
                    sys.stdout.flush()
            if ord3 == 5:
                sys.stdout.write(" \\\\ \\hline \\hline\n")
            else:
                sys.stdout.write(" \\\\ \\hline\n")


def findRepresentative(R, ord2, ord3, unit2Residue, unit3Residue, sign):
    unit2Residue = int(unit2Residue)
    unit3Residue = int(unit3Residue)
    step = 2^ord2 * 3^ord3
    for test in range(sign* step, sign * 2^8 * 3^8, sign * step):
        if test % 2^(ord2+1) != 0:
            if test % 3^(ord3+1) != 0:
                unitAt2 = test / 2^ord2
                unitAt3 = test / 3^ord3
                # this could be optimized
                if unitAt2 % 2^3 == unit2Residue:
                    if unitAt3 % 3^3  == unit3Residue:
                        return test
    raise ValueError("No representative below 2^8 * 3^8")


def getResiduesModPower(ideal, power):
    residues = list(ideal.invertible_residues())
    residuesToPower = []
    for elem in residues:
        if ideal.reduce(elem^power) not in residuesToPower:
            residuesToPower.append(ideal.reduce(elem^power))
    return list(ideal.invertible_residues_mod(residuesToPower))


# Returns legendre(-3, U(d)), in the notation of the main thm.
# Right now this only works when 6 is stationary in K.
def legendreCondition(K, d):
    ans = 1
    factors = K.factor(d)
    for factor in factors:
        if factor[1] % 6 == 2 or factor[1] % 6 == 4:
            if factor[0] != 2 and factor[0] != 3:
                ans *= factor[0].residue_symbol(-3, 2)
    return ans


def getConjecturalRankParity(E):
    twoSelmerRankParity = E.simon_two_descent()[1] % 2
    torsion = E.torsion_subgroup()
    twoTorsionCount = 0
    for g in torsion:
        if g + g == 0:
            twoTorsionCount += 1
    twoTorsionRank = log(twoTorsionCount, 2)
    return (twoSelmerRankParity + twoTorsionRank) % 2


classifySexticTwistsOverQTable(1)
classifySexticTwistsOverQTable(-1)
